/*
    Copyright 2006,2007,2008 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <wincrypt.h>



typedef unsigned char   u_char;



#define VER         "0.1.1"
#define PRESSKEY    printf("\n  Press RETURN to exit\n"); \
                    fgetc(stdin);



u_char *find_pwd(u_char *data, u_char *pwd_pattern);
u_char *myfgets(u_char *buff, int size, FILE *fd, int w);
int decshow(u_char *data, u_char *entropy, int rawdata);
void wcheck(u_char *data);
void hex2c(u_char *in, DATA_BLOB *out);



int main(int argc, char *argv[]) {
    FILE    *fd;
    int     w   = 0,    // used for unicode
            dec = 0,
            len;
    u_char  buff[8192],
            *data,
            *entropy = NULL,
            *p;

    fputs("\n"
        "Generic CryptUnprotectData and RDP passwords decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <hash/any_file/file.rdp> [entropy]\n"
            "\n"
            "The tool automatically searches all the password fields in the file or\n"
            "decrypts directly the command-line argument if it's a hash\n"
            "entropy is a possible additional hash (used as password) if required\n"
            "Note that ONLY the user who has encrypted the password can decrypt it\n"
            "\n", argv[0]);
        PRESSKEY;
        exit(1);
    }

    data = strdup(argv[1]);
    if(argc > 2) entropy = strdup(argv[2]);

    if(!strcmp(data, "-")) {
        printf("- input file:  %s\n", "stdin");
        fd = stdin;
        // no unicode check on stdin since is not possible to rewind
    } else {
        fd = fopen(data, "rb");
        if(!fd) {
            printf("- input is the encrypted string\n");
            decshow(data, entropy, 0);
            return(0);
        }
        printf("- input file:  %s\n", data);
        fread(buff, 2, 1, fd);
        if((buff[0] == 0xff) && (buff[1] == 0xfe)) {
            w = 1;
        } else {
            rewind(fd);
        }
    }

    while(myfgets(buff, sizeof(buff), fd, w)) {
        p = find_pwd(buff, "pass");
        if(!p) continue;
        if(!decshow(p, entropy, 0)) dec++;
    }

    if(!dec) {
        fseek(fd, w ? 2 : 0, SEEK_SET);
        myfgets(buff, sizeof(buff), fd, w);
        if(decshow(buff, entropy, 0) < 0) {
            printf("- no password found, I consider the input as raw data instead of hash\n");
            rewind(fd);
            len = fread(buff, 1, sizeof(buff), fd);
            decshow(buff, entropy, len);
        }
    }

    fclose(fd);
    PRESSKEY;
    return(0);
}



u_char *find_pwd(u_char *data, u_char *pwd_pattern) {
    u_char  *p,
            *l;

    for(p = data; *p && (*p < ' '); p++);                               // find parameter
    if(!*p) return(NULL);

    for(l = p; *l && (*l > ' ') && (*l != '=') && (*l != ':'); l++);    // delimit par
    if(!*l) return(NULL);
    *l = 0;

    if(!stristr(p, pwd_pattern)) return(NULL);

    l++;
    p = strchr(l, '=');         // password=hash
    if(p) {
        for(p++; *p && (*p < ' '); p++);
        if(!*p) return(NULL);
    } else {
        p = strrchr(l, ':');    // password 51:b:hash
        if(!p) return(l);
        p++;
    }

    return(p);
}



u_char *myfgets(u_char *buff, int size, FILE *fd, int w) {
    int     t;
    u_char  *p,
            *l;

    p = buff;
    l = buff + size - 1;
    while(p < l) {
        t = fgetc(fd);
        if(t < 0) {
            if(p == buff) return(NULL);
            break;
        }
        if(w) fgetc(fd);
        if(t == '\r') continue; // skip \r
        if(t == '\n') break;
        *p++ = t;
    }
    *p = 0;
    return(buff);
}



int decshow(u_char *data, u_char *entropy, int rawdata) {
    DATA_BLOB   DataIn,
                DataEntropy,
                DataOut;
    LPWSTR      description;

    memset(&DataIn,      0, sizeof(DataIn));
    memset(&DataEntropy, 0, sizeof(DataEntropy));
    memset(&DataOut,     0, sizeof(DataOut));
    description = NULL;

    if(!rawdata && (!strnicmp(data, "51:b:", 5) || !strnicmp(data, "x9392", 5))) {
        data += 5;
    }

    if(!rawdata) {
        hex2c(data, &DataIn);
    } else {
        DataIn.pbData = LocalAlloc(LPTR, rawdata);
        if(!DataIn.pbData) return(-1);
        memcpy(DataIn.pbData, data, rawdata);
        DataIn.cbData = rawdata;
    }
    if(entropy) hex2c(data, &DataEntropy);

    if(!CryptUnprotectData(
      &DataIn,                          // input data
      &description,                     // output description
      entropy ? &DataEntropy : NULL,    // optional entropy
      NULL,                             // reserved
      NULL,                             // optional prompt structure
      CRYPTPROTECT_UI_FORBIDDEN,        // flags
      &DataOut)) {                      // output
        printf(
            "- decryption error, the input is invalid or you can't decrypt it because the\n"
            "  data has been encrypted on another system or with another account\n");
        return(-1);
    }

    wcheck(DataOut.pbData);
    printf("\n- decrypted:   %s\n", DataOut.pbData);
    if(description) {
        wcheck((u_char *)description);
        printf("- description: %s\n", (char *)description);
        LocalFree(description);
    }

    if(DataOut.pbData)      LocalFree(DataOut.pbData);
    if(DataIn.pbData)       LocalFree(DataIn.pbData);
    if(DataEntropy.pbData)  LocalFree(DataEntropy.pbData);
    return(0);
}



void wcheck(u_char *data) {
    u_char  *p;

    if(data[0] && data[1]) return;     // blind if no unicode and one byte only!

    for(p = data; *data; data += 2, p++) {
        *p = *data;
    }
    *p = 0;
}



void hex2c(u_char *in, DATA_BLOB *out) {
    int     tmp;
    u_char  *p;

    tmp = strlen(in) / 2;
    out->pbData = LocalAlloc(LPTR, tmp);
    if(!out->pbData) {
        out->cbData = 0;
        return;
    }
    p = out->pbData;

    while(in[0] && in[1] && (sscanf(in, "%02x", &tmp) == 1)) {
        *p++ = tmp;
        in += 2;
    }
    out->cbData = p - out->pbData;
}


